/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022                        ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include <pecunia/Codes.h>
using pecunia::currency::Iso4217Codes;
#include <pecunia/Math.h>
using pecunia::math::sum;
#include <pecunia/Money.h>
using pecunia::currency::Money;
#include <pecunia/MoneyTypes.h>
using pecunia::MinorSign;
#include <QDate>
using ::QDate;
#include <QObject>
using ::QObject;
// using Q_OBJECT
#include <QTest>
// using QTEST_MAIN
// using QVERIFY_EXCEPTION_THROWN
#include <qt-mocks/Matchers.h>
using qt_mocks::anyOf;
using qt_mocks::equals;
#include <qt-mocks/Verify.hpp>
// using QVERIFY_FALSE
// using QVERIFY_THAT
// using QVERIFY_TRUE

#include <algorithm>
using std::transform;
#include <chrono>
using namespace std::chrono_literals;
#include <map>
using std::map;
#include <set>
using std::set;
#include <iterator>
using std::inserter;
using std::back_inserter;
#include <vector>
using std::vector;

#include <accounting/AccountCode.h>
using drn::accounting::AccountCode;
#include <accounting/AccountingErrors.h>
using drn::accounting::AccountError;
#include <accounting/AccountNumber.h>
using drn::accounting::AccountNumber;
#include <accounting/LedgerBooks.h>
using drn::accounting::LedgerAccount;
#include <accounting/Transaction.h>
using drn::accounting::Transaction;
using drn::accounting::TransactionNumber;
using drn::accounting::TransactionStatuses;
#include <banking/Bank.h>
using drn::banking::Bank;
#include <banking/BankingErrors.h>
using drn::banking::BankError;
#include <banking/BankAccount.h>
using drn::banking::BankAccount;
#include <banking/BankAccountType.h>
using drn::banking::BankAccountType;
#include <banking/BankName.h>
using drn::banking::BankName;
#include <banking/BankAccount.h>
using drn::banking::SupportedAccountTypes;
#include <budgeting/BudgetItemDue.h>
using drn::budgeting::BudgetItemDue;
#include <budgeting/BudgetItemIdentifier.h>
using drn::budgeting::BudgetItemIdentifier;
#include <budgeting/BudgetItemTypes.h>
using drn::budgeting::BudgetItemTypes;
#include <budgeting/BudgetSource.h>
using drn::budgeting::BudgetSource;
#include <budgeting/Bill.h>
using drn::budgeting::Bill;
#include <budgeting/EventFrequency.h>
using drn::budgeting::EventFrequency;
#include <budgeting/Wage.h>
using drn::budgeting::Wage;
#include <foundation/CalendarTime.hpp>
using drn::foundation::Days;
#include <foundation/Optional.hpp>
using drn::foundation::Optional;
#include <foundation/PresentationText.h>
using ::operator<<;
#include <internal/BudgetBankLedgers.h>
using drn::navigation::internal::BudgetBankLedgers;
#include <surveying/BankAccountSurvey.h>
using drn::surveying::BankAccountSurvey;


namespace drn
{
namespace navigation
{
namespace tests
{

struct ProjectedSurveyUT : QObject
{
	Q_OBJECT

	BudgetBankLedgers budgetBankLedger_{};
	const QDate surveyOn_{QDate::currentDate()};
	const QDate oneMonthAndDayBeforeSurveyOn_{this->surveyOn_.addMonths(-1).addDays(-1)};
	const QDate surveyYearOn_{this->surveyOn_.addMonths(11)};
	const Bill clothing_{
		BudgetSource{"Clothing"},
		Money{50, 0u, Iso4217Codes::USD},
		EventFrequency::Monthly,
		this->oneMonthAndDayBeforeSurveyOn_
	};
	const BudgetItemIdentifier clothingId_{BudgetItemTypes::Bill, this->clothing_.source()};
	AccountCode clothingAccount_{};
	BankAccount bankAccount_{
		BankName{"Community 1st"},
		BankAccountType{{}, SupportedAccountTypes::Chequeing}
	};
	const Wage salary_{
		BudgetSource{"Company Salary"},
		Money{100, 0u, Iso4217Codes::USD},
		EventFrequency::Monthly,
		this->oneMonthAndDayBeforeSurveyOn_
	};
	const BudgetItemIdentifier salaryId_{BudgetItemTypes::Wage, this->salary_.source()};
	AccountCode salaryAccount_{};
	const QDate oneYearAndDayBeforeSurveyOn_{this->surveyYearOn_.addDays(-1)};
	const Bill streaming_{
		BudgetSource{"Video Streaming"},
		Money{120, 0u, Iso4217Codes::USD},
		EventFrequency::Annually,
		this->oneYearAndDayBeforeSurveyOn_
	};
	const BudgetItemIdentifier streamingId_{BudgetItemTypes::Bill, this->streaming_.source()};

	set<Transaction> simulateBillPosting(
		const qint32 numberMonths,
		const BudgetItemIdentifier& id,
		const BudgetSource& source,
		const Money& amount,
		const QDate& dueOn
	)
	{
		set<TransactionNumber> allPosted{};

		for (auto count{0} ; count < numberMonths ; ++count)
		{
			const BudgetItemDue due{id, amount, dueOn.addMonths(count)};
			allPosted.emplace(
				this->budgetBankLedger_.post(
					due,
					this->bankAccount_.account_.code_,
					dueOn.addMonths(count)
				)
			);
		}
		set<Transaction> txns;
		const auto ledgerTxns{
			this->budgetBankLedger_.generalLedger().ledger(
				this->budgetBankLedger_.budgetLedgers().budgetAccountCode<Bill>(source)
			).transactions_
		};

		for (const auto& posted : allPosted)
			txns.emplace(ledgerTxns.at(posted));
		return txns;
	}

	void simulateBillSkipping(const qint32 numberMonths, const BudgetItemIdentifier& id)
	{
		for (auto count{0} ; count < numberMonths ; ++count)
			this->budgetBankLedger_.skip(id);
	}

private slots:
	void init()
	{
		this->salaryAccount_ = this->budgetBankLedger_.add(this->salary_);
		this->clothingAccount_ = this->budgetBankLedger_.add(this->clothing_);
		this->budgetBankLedger_.add(Bank{this->bankAccount_.bank_});
		this->bankAccount_.account_.code_ = this->budgetBankLedger_.addAccount(
			"Current",
			this->bankAccount_.account_.type_,
			Money{Iso4217Codes::USD}
		);
		this->budgetBankLedger_.addAssociation(
			this->bankAccount_.bank_,
			this->bankAccount_.account_.code_.number(),
			this->bankAccount_.account_.type_
		);
		this->budgetBankLedger_.post(
			BudgetItemDue{
				this->salaryId_,
				this->salary_.amount(),
				this->oneMonthAndDayBeforeSurveyOn_
			},
			this->bankAccount_.account_.code_,
			this->oneMonthAndDayBeforeSurveyOn_
		);
	}

	void cleanup()
	{
		this->budgetBankLedger_ = {};
	}

	void projectedSurvey_bankDoesNotExist_ShouldThrow()
	{
		QVERIFY_EXCEPTION_THROWN(
			this->budgetBankLedger_.projectSurvey(
				BankName{"Fake Bank"},
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_,
				40h,
				Days{5}
			),
			BankError
		);
	}

	void projectedSurvey_accountNumberDoesNotExist_ShouldThrow()
	{
		QVERIFY_EXCEPTION_THROWN(
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				AccountNumber{10001},
				this->surveyOn_,
				40h,
				Days{5}
			),
			AccountError
		);
	}

	void projectedSurvey_postedWithReconciledTransaction_ShouldNotHaveReconciledTransactions()
	{
		const auto txns{
			this->simulateBillPosting(
				1,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn()
			)
		};
		this->budgetBankLedger_.reconcile(*txns.cbegin()->number());
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_,
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.usedBudget().count(this->clothingId_), equals(size_t{0ul}));
	}

	void projectedSurvey_postedWithFutureTransaction_ShouldNotHaveFutureTransactionsButUnspentAssumedAllItems()
	{
		const auto txns{
			this->simulateBillPosting(
				1,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->oneMonthAndDayBeforeSurveyOn_,
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.usedBudget().count(this->clothingId_), equals(size_t{0ul}));
		QVERIFY_THAT(
			survey.unusedBudget(),
			equals(
				map<BudgetItemIdentifier, Money>{
					{
						{this->salaryId_, Money{Iso4217Codes::USD}},
						{this->clothingId_, Money{Iso4217Codes::USD}}
					}
				}
			)
		);
	}

	//{ Bill Tests

	void bill_1month_skipped_ShouldBeCalculated()
	{
		this->simulateBillSkipping(1, this->clothingId_);
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_,
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.reconciled().bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(survey.reconciled().balance(), equals(Money{Iso4217Codes::USD}));
		QVERIFY_THAT(survey.projectedBalance(), equals(this->salary_.amount()));
		QVERIFY_FALSE(survey.reconciled().reconciledOn().hasValue());
		QVERIFY_THAT(survey.usedBudget().size(), equals(static_cast<size_t>(1)));
		QVERIFY_THAT(survey.usedBudget().cbegin()->second.size(), equals(static_cast<size_t>(1)));
		const auto& actualTxn{*survey.usedBudget().cbegin()->second.cbegin()};
		const Transaction expectedTxn{
			this->oneMonthAndDayBeforeSurveyOn_,
			this->bankAccount_.account_.code_,
			this->salaryAccount_,
			this->salary_.amount(),
			QStringLiteral("Some description"),
			TransactionNumber{1u},
			TransactionStatuses::Unreconciled
		};
		QVERIFY_THAT(actualTxn.occurredOn(), equals(expectedTxn.occurredOn()));
		QVERIFY_THAT(actualTxn.debiting(), equals(expectedTxn.debiting()));
		QVERIFY_THAT(actualTxn.crediting(), equals(expectedTxn.crediting()));
		QVERIFY_THAT(actualTxn.amount(), equals(expectedTxn.amount()));
		QVERIFY_THAT(actualTxn.number(), equals(expectedTxn.number()));
		QVERIFY_THAT(actualTxn.status(), equals(expectedTxn.status()));
		QVERIFY_THAT(
			survey.unusedBudget(),
			equals(
				map<BudgetItemIdentifier, Money>{
					{
						{this->salaryId_, Money{Iso4217Codes::USD}},
						{this->clothingId_, this->clothing_.amount()}
					}
				}
			)
		);
	}

	void bill_1month_postedExact_ShouldBeCalculated()
	{
		const auto txns{
			this->simulateBillPosting(
				1,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_,
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.reconciled().bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(survey.reconciled().balance(), equals(Money{Iso4217Codes::USD}));
		QVERIFY_THAT(survey.projectedBalance(), equals(Money{50, 0u, Iso4217Codes::USD}));
		QVERIFY_FALSE(survey.reconciled().reconciledOn().hasValue());
		QVERIFY_THAT(survey.usedBudget().at(this->clothingId_), equals(txns));
		QVERIFY_THAT(
			survey.unusedBudget(),
			equals(
				map<BudgetItemIdentifier, Money>{
					{
						{this->salaryId_, Money{Iso4217Codes::USD}},
						{this->clothingId_, Money{Iso4217Codes::USD}}
					}
				}
			)
		);
	}

	void bill_1month_postedUnder_ShouldBeCalculated()
	{
		const auto txns{
			this->simulateBillPosting(
				1,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount() - Money{5, 0u, Iso4217Codes::USD},
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_,
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.reconciled().bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(survey.reconciled().balance(), equals(Money{Iso4217Codes::USD}));
		QVERIFY_THAT(survey.projectedBalance(), equals(Money{55, 0u, Iso4217Codes::USD}));
		QVERIFY_FALSE(survey.reconciled().reconciledOn().hasValue());
		QVERIFY_THAT(survey.usedBudget().at(this->clothingId_), equals(txns));
		QVERIFY_THAT(
			survey.unusedBudget(),
			equals(
				map<BudgetItemIdentifier, Money>{
					{
						{this->salaryId_, Money{Iso4217Codes::USD}},
						{this->clothingId_, Money{5, 0u, Iso4217Codes::USD}}
					}
				}
			)
		);
	}

	void bill_1month_postedOver_ShouldBeCalculated()
	{
		const auto txns{
			this->simulateBillPosting(
				1,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount() + Money{10, 0u, Iso4217Codes::USD},
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_,
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.reconciled().bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(survey.reconciled().balance(), equals(Money{Iso4217Codes::USD}));
		QVERIFY_THAT(survey.projectedBalance(), equals(Money{40, 0u, Iso4217Codes::USD}));
		QVERIFY_FALSE(survey.reconciled().reconciledOn().hasValue());
		QVERIFY_THAT(survey.usedBudget().at(this->clothingId_), equals(txns));
		QVERIFY_THAT(
			survey.unusedBudget(),
			equals(
				map<BudgetItemIdentifier, Money>{
					{
						{this->salaryId_, Money{Iso4217Codes::USD}},
						{this->clothingId_, Money{-10, 0u, Iso4217Codes::USD}}
					}
				}
			)
		);
	}

	void bill_1month_postedAnnualExact_ShouldBeCalculated()
	{
		this->budgetBankLedger_.add(this->streaming_);
		const auto txns{
			this->simulateBillPosting(
				1,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_,
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.reconciled().bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(survey.reconciled().balance(), equals(Money{Iso4217Codes::USD}));
		QVERIFY_THAT(survey.projectedBalance(), equals(Money{50, 0u, Iso4217Codes::USD}));
		QVERIFY_FALSE(survey.reconciled().reconciledOn().hasValue());
		QVERIFY_THAT(survey.usedBudget().at(this->clothingId_), equals(txns));
		QVERIFY_THAT(
			survey.unusedBudget(),
			equals(
				map<BudgetItemIdentifier, Money>{
					{
						{this->salaryId_, Money{Iso4217Codes::USD}},
						{this->streamingId_, Money{9, 9999u, Iso4217Codes::USD}},
						{this->clothingId_, Money{this->clothing_.amount().code()}}
					}
				}
			)
		);
	}

	void bill_3month_postedExact_ShouldBeCalculated()
	{
		const auto txns{
			this->simulateBillPosting(
				3,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_.addMonths(2),
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.reconciled().bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(survey.reconciled().balance(), equals(Money{Iso4217Codes::USD}));
		QVERIFY_THAT(survey.projectedBalance(), equals(Money{-50, 0u, Iso4217Codes::USD}));
		QVERIFY_FALSE(survey.reconciled().reconciledOn().hasValue());
		QVERIFY_THAT(survey.usedBudget().at(this->clothingId_), equals(txns));
		QVERIFY_THAT(
			survey.unusedBudget(),
			equals(
				map<BudgetItemIdentifier, Money>{
					{
						{this->salaryId_, this->salary_.amount() * 2}, // Skipped 2 months
						{this->clothingId_, Money{Iso4217Codes::USD}}
					}
				}
			)
		);
	}

	void bill_3month_postedUnder_ShouldBeCalculated()
	{
		const auto txns{
			this->simulateBillPosting(
				3,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount() - Money{5, 0u, Iso4217Codes::USD},
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_.addMonths(2),
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.reconciled().bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(survey.reconciled().balance(), equals(Money{Iso4217Codes::USD}));
		QVERIFY_THAT(survey.projectedBalance(), equals(Money{-35, 0u, Iso4217Codes::USD}));
		QVERIFY_FALSE(survey.reconciled().reconciledOn().hasValue());
		QVERIFY_THAT(survey.usedBudget().at(this->clothingId_), equals(txns));
		QVERIFY_THAT(
			survey.unusedBudget(),
			equals(
				map<BudgetItemIdentifier, Money>{
					{
						{this->salaryId_, this->salary_.amount() * 2}, // Skipped 2 months
						{this->clothingId_, Money{15, 0u, Iso4217Codes::USD}}
					}
				}
			)
		);
	}

	void bill_3month_postedOver_ShouldBeCalculated()
	{
		const auto txns{
			this->simulateBillPosting(
				3,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount() + Money{10, 0u, Iso4217Codes::USD},
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_.addMonths(2),
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.reconciled().bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(survey.reconciled().balance(), equals(Money{Iso4217Codes::USD}));
		QVERIFY_THAT(survey.projectedBalance(), equals(Money{-80, 0u, Iso4217Codes::USD}));
		QVERIFY_FALSE(survey.reconciled().reconciledOn().hasValue());
		QVERIFY_THAT(survey.usedBudget().at(this->clothingId_), equals(txns));
		QVERIFY_THAT(
			survey.unusedBudget(),
			equals(
				map<BudgetItemIdentifier, Money>{
					{
						{this->salaryId_, this->salary_.amount() * 2}, // Skipped 2 months
						{this->clothingId_, Money{-30, 0u, Iso4217Codes::USD}}
					}
				}
			)
		);
	}

	void bill_3month_postedAnnualExact_ShouldBeCalculated()
	{
		this->budgetBankLedger_.add(this->streaming_);
		const auto monthlyTxns{
			this->simulateBillPosting(
				3,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_.addMonths(2),
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.reconciled().bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(survey.reconciled().balance(), equals(Money{Iso4217Codes::USD}));
		QVERIFY_THAT(survey.projectedBalance(), equals(Money{-50, 0u, Iso4217Codes::USD}));
		QVERIFY_FALSE(survey.reconciled().reconciledOn().hasValue());
		QVERIFY_THAT(survey.usedBudget().at(this->clothingId_), equals(monthlyTxns));
		QVERIFY_THAT(
			survey.unusedBudget(),
			equals(
				map<BudgetItemIdentifier, Money>{
					{
						{this->salaryId_, this->salary_.amount() * 2}, // Skipped 2 months
						{this->streamingId_, Money{29, 9997u, Iso4217Codes::USD}},
						{this->clothingId_, Money{this->clothing_.amount().code()}}
					}
				}
			)
		);
	}

	void bill_12month_postedExact_ShouldBeCalculated()
	{
		const auto txns{
			this->simulateBillPosting(
				12,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyYearOn_,
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.reconciled().bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(survey.reconciled().balance(), equals(Money{Iso4217Codes::USD}));
		QVERIFY_THAT(survey.projectedBalance(), equals(Money{-500, 0u, Iso4217Codes::USD}));
		QVERIFY_FALSE(survey.reconciled().reconciledOn().hasValue());
		QVERIFY_THAT(survey.usedBudget().at(this->clothingId_), equals(txns));
		QVERIFY_THAT(
			survey.unusedBudget(),
			equals(
				map<BudgetItemIdentifier, Money>{
					{
						{this->salaryId_, this->salary_.amount() * 11}, // Skipped 11 months.
						{this->clothingId_, Money{Iso4217Codes::USD}}
					}
				}
			)
		);
	}

	void bill_12month_postedUnder_ShouldBeCalculated()
	{
		const auto txns{
			this->simulateBillPosting(
				12,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount() - Money{5, 0u, Iso4217Codes::USD},
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyYearOn_,
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.reconciled().bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(survey.reconciled().balance(), equals(Money{Iso4217Codes::USD}));
		QVERIFY_THAT(survey.projectedBalance(), equals(Money{-440, 0u, Iso4217Codes::USD}));
		QVERIFY_FALSE(survey.reconciled().reconciledOn().hasValue());
		QVERIFY_THAT(survey.usedBudget().at(this->clothingId_), equals(txns));
		QVERIFY_THAT(
			survey.unusedBudget(),
			equals(
				map<BudgetItemIdentifier, Money>{
					{
						{this->salaryId_, this->salary_.amount() * 11}, // Skipped 11 months
						{this->clothingId_, Money{60, 0u, Iso4217Codes::USD}}
					}
				}
			)
		);
	}

	void bill_12month_postedOver_ShouldBeCalculated()
	{
		const auto txns{
			this->simulateBillPosting(
				12,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount() + Money{10, 0u, Iso4217Codes::USD},
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyYearOn_,
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.reconciled().bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(survey.reconciled().balance(), equals(Money{Iso4217Codes::USD}));
		QVERIFY_THAT(survey.projectedBalance(), equals(Money{-620, 0u, Iso4217Codes::USD}));
		QVERIFY_FALSE(survey.reconciled().reconciledOn().hasValue());
		QVERIFY_THAT(survey.usedBudget().at(this->clothingId_), equals(txns));
		QVERIFY_THAT(
			survey.unusedBudget(),
			equals(
				map<BudgetItemIdentifier, Money>{
					{
						{this->salaryId_, this->salary_.amount() * 11}, // Skipped 11 months
						{this->clothingId_, Money{-120, 0u, Iso4217Codes::USD}}
					}
				}
			)
		);
	}

	void bill_12month_postedAnnualExact_ShouldBeCalculated()
	{
		this->budgetBankLedger_.add(this->streaming_);
		const auto monthlyTxns{
			this->simulateBillPosting(
				12,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn()
			)
		};
		const auto annualTxns{
			this->simulateBillPosting(
				1,
				this->streamingId_,
				this->streaming_.source(),
				this->streaming_.amount(),
				this->streaming_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyYearOn_,
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.reconciled().bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(survey.reconciled().balance(), equals(Money{Iso4217Codes::USD}));
		QVERIFY_THAT(
			this->budgetBankLedger_.generalLedger()
				.ledger(this->bankAccount_.account_.code_).account_.balance(),
			equals(Money{-620, 0u, Iso4217Codes::USD})
		);
		QVERIFY_THAT(survey.projectedBalance(), equals(Money{-620, 0u, Iso4217Codes::USD}));
		QVERIFY_FALSE(survey.reconciled().reconciledOn().hasValue());
		QVERIFY_THAT(survey.usedBudget().at(this->clothingId_), equals(monthlyTxns));
		QVERIFY_THAT(survey.usedBudget().at(this->streamingId_), equals(annualTxns));
		QVERIFY_THAT(
			survey.unusedBudget(),
			equals(
				map<BudgetItemIdentifier, Money>{
					{
						{this->salaryId_, this->salary_.amount() * 11}, // Skipped 11 months
						{
							this->streamingId_,
							Money{MinorSign::Negative, 12, this->streaming_.amount().code()}
						},
						{this->clothingId_, Money{this->clothing_.amount().code()}}
					}
				}
			)
		);
	}

	void bill_13month_annualPaidSurveyedNextMonth_ShouldBeCalculated()
	{
		this->budgetBankLedger_.add(this->streaming_);
		const auto monthlyTxns{
			this->simulateBillPosting(
				13,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn()
			)
		};
		const auto annualTxns{
			this->simulateBillPosting(
				1,
				this->streamingId_,
				this->streaming_.source(),
				this->streaming_.amount(),
				this->streaming_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_.addMonths(12),
				40h,
				Days{5}
			)
		};
		QVERIFY_THAT(survey.reconciled().bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(survey.reconciled().balance(), equals(Money{Iso4217Codes::USD}));
		QVERIFY_THAT(survey.projectedBalance(), equals(Money{-670, 0u, Iso4217Codes::USD}));
		QVERIFY_FALSE(survey.reconciled().reconciledOn().hasValue());
		QVERIFY_THAT(survey.usedBudget().at(this->clothingId_), equals(monthlyTxns));
		QVERIFY_THAT(survey.usedBudget().at(this->streamingId_), equals(annualTxns));
		QVERIFY_THAT(
			survey.unusedBudget(),
			equals(
				map<BudgetItemIdentifier, Money>{
					{
						{this->salaryId_, this->salary_.amount() * 12}, // Skipped 12 months
						{this->streamingId_, Money{9, 9987u, Iso4217Codes::USD}},
						{this->clothingId_, Money{this->clothing_.amount().code()}}
					}
				}
			)
		);
	}

	//}

	//{

	void bill_completedSurveyFirstTime_ShouldBeReconciled()
	{
		this->budgetBankLedger_.add(this->streaming_);
		const auto txns{
			this->simulateBillPosting(
				1,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_,
				40h,
				Days{5}
			)
		};
		vector<TransactionNumber> expectedReconciled{
			{
				this->budgetBankLedger_.generalLedger().ledger(this->salaryAccount_)
					.transactions_.begin()->first
			}
		};
		transform(
			txns.cbegin(),
			txns.cend(),
			back_inserter(expectedReconciled),
			[] (const auto& txn)
			{
				return *txn.number();
			}
		);
		const vector<Transaction> expectedAdded{};
		map<BudgetItemIdentifier, Money> expectedUnusedBudget{
			{
				{this->salaryId_, Money{Iso4217Codes::USD}},
				{this->clothingId_, Money{Iso4217Codes::USD}},
				{this->streamingId_, Money{50, 0u, Iso4217Codes::USD}}
			}
		};
		const auto& bankAccount{survey.reconciled().bankAccount()};
		QVERIFY_NO_THROW(
			this->budgetBankLedger_.completedSurvey(
				this->surveyOn_,
				bankAccount,
				expectedReconciled,
				expectedAdded,
				expectedUnusedBudget
			)
		);
		const auto& surveys{this->budgetBankLedger_.surveys().at(bankAccount)};
		const auto& reconciliation{surveys.reconciled()};
		QVERIFY_THAT(reconciliation.reconciledOn(), equals(Optional<QDate>{this->surveyOn_}));
		QVERIFY_THAT(reconciliation.balance(), equals(Money{50, 0u, Iso4217Codes::USD}));

		for (const auto& codeLedgerAccount : this->budgetBankLedger_.generalLedger().ledgers())
		{
			const auto& accountCode{codeLedgerAccount.first};
			const auto& ledgerAccount{codeLedgerAccount.second};

			for (const auto& txnNumberTxn : codeLedgerAccount.second.transactions_)
			{
				const auto& txn{txnNumberTxn.second};
				QVERIFY_THAT(txn.status(), equals(TransactionStatuses::Reconciled));

				if (accountCode == this->bankAccount_.account_.code_)
				{
					const auto totalUnused{
						sum(
							expectedUnusedBudget.begin()->second.code(),
							expectedUnusedBudget
						)
					};
					QVERIFY_THAT(ledgerAccount.account_.balance(), equals(totalUnused));
				}
			}
		}
	}
	void bill_completedReconciledTxnMismatch_ShouldThrow()
	{
		this->budgetBankLedger_.add(this->streaming_);
		const auto txns{
			this->simulateBillPosting(
				1,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_,
				40h,
				Days{5}
			)
		};
		vector<TransactionNumber> expectedReconciled{};
		transform(
			txns.cbegin(),
			txns.cend(),
			back_inserter(expectedReconciled),
			[] (const auto& txn)
			{
				return *txn.number();
			}
		);
		const vector<Transaction> expectedAdded{};
		map<BudgetItemIdentifier, Money> expectedUnusedBudget{
			{
				{this->salaryId_, Money{Iso4217Codes::USD}},
				{this->clothingId_, Money{Iso4217Codes::USD}},
				{this->streamingId_, Money{50, 0u, Iso4217Codes::USD}}
			}
		};
		const auto& bankAccount{survey.reconciled().bankAccount()};
		const auto originalBalance{
			this->budgetBankLedger_.generalLedger().ledger(this->bankAccount_.account_.code_).account_
				.balance()
		};
		QVERIFY_EXCEPTION_THROWN(
			this->budgetBankLedger_.completedSurvey(
				this->surveyOn_,
				bankAccount,
				expectedReconciled,
				expectedAdded,
				expectedUnusedBudget
			),
			BankError
		);
		QVERIFY_THAT(
			this->budgetBankLedger_.surveys().count(bankAccount),
			equals(static_cast<size_t>(0))
		);

		for (const auto& codeLedgerAccount : this->budgetBankLedger_.generalLedger().ledgers())
		{
			const auto& accountCode{codeLedgerAccount.first};
			const auto& ledgerAccount{codeLedgerAccount.second};

			for (const auto& txnNumberTxn : codeLedgerAccount.second.transactions_)
			{
				const auto& txn{txnNumberTxn.second};
				QVERIFY_THAT(txn.status(), equals(TransactionStatuses::Unreconciled));

				if (accountCode == this->bankAccount_.account_.code_)
					QVERIFY_THAT(ledgerAccount.account_.balance(), equals(originalBalance));
			}
		}
	}

	void bill_completedSurveyUnusedBalanceMismatch_ShouldThrow()
	{
		this->budgetBankLedger_.add(this->streaming_);
		const auto txns{
			this->simulateBillPosting(
				1,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_,
				40h,
				Days{5}
			)
		};
		vector<TransactionNumber> expectedReconciled{
			{
				this->budgetBankLedger_.generalLedger().ledger(this->salaryAccount_)
					.transactions_.begin()->first
			}
		};
		transform(
			txns.cbegin(),
			txns.cend(),
			back_inserter(expectedReconciled),
			[] (const auto& txn)
			{
				return *txn.number();
			}
		);
		const vector<Transaction> expectedAdded{};
		map<BudgetItemIdentifier, Money> expectedUnusedBudget{
			{
				{this->salaryId_, Money{Iso4217Codes::USD}},
				{this->clothingId_, Money{Iso4217Codes::USD}},
				{this->streamingId_, Money{Iso4217Codes::USD}}
			}
		};
		const auto& bankAccount{survey.reconciled().bankAccount()};
		const auto originalBalance{
			this->budgetBankLedger_.generalLedger().ledger(this->bankAccount_.account_.code_).account_
				.balance()
		};
		QVERIFY_EXCEPTION_THROWN(
			this->budgetBankLedger_.completedSurvey(
				this->surveyOn_,
				bankAccount,
				expectedReconciled,
				expectedAdded,
				expectedUnusedBudget
			),
			BankError
		);
		QVERIFY_THAT(
			this->budgetBankLedger_.surveys().count(bankAccount),
			equals(static_cast<size_t>(0))
		);

		for (const auto& codeLedgerAccount : this->budgetBankLedger_.generalLedger().ledgers())
		{
			const auto& accountCode{codeLedgerAccount.first};
			const auto& ledgerAccount{codeLedgerAccount.second};

			for (const auto& txnNumberTxn : codeLedgerAccount.second.transactions_)
			{
				const auto& txn{txnNumberTxn.second};
				QVERIFY_THAT(txn.status(), equals(TransactionStatuses::Unreconciled));

				if (accountCode == this->bankAccount_.account_.code_)
					QVERIFY_THAT(ledgerAccount.account_.balance(), equals(originalBalance));
			}
		}
	}

	void bill_completedSurveyAdditionalExpenseMismatch_ShouldThrow()
	{
		this->budgetBankLedger_.add(this->streaming_);
		const auto txns{
			this->simulateBillPosting(
				1,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_,
				40h,
				Days{5}
			)
		};
		vector<TransactionNumber> expectedReconciled{
			{
				this->budgetBankLedger_.generalLedger().ledger(this->salaryAccount_)
					.transactions_.begin()->first
			}
		};
		transform(
			txns.cbegin(),
			txns.cend(),
			back_inserter(expectedReconciled),
			[] (const auto& txn)
			{
				return *txn.number();
			}
		);
		const vector<Transaction> expectedAdded{
			{
				{
					this->surveyOn_,
					this->clothingAccount_,
					this->bankAccount_.account_.code_,
					this->clothing_.amount(),
					QStringLiteral("Additional missed transaction.")
				}
			}
		};
		map<BudgetItemIdentifier, Money> expectedUnusedBudget{
			{
				{this->salaryId_, Money{Iso4217Codes::USD}},
				{this->clothingId_, Money{Iso4217Codes::USD}},
				{this->streamingId_, Money{50, 0u, Iso4217Codes::USD}}
			}
		};
		const auto& bankAccount{survey.reconciled().bankAccount()};
		const auto originalBalance{
			this->budgetBankLedger_.generalLedger().ledger(this->bankAccount_.account_.code_).account_
			.balance()
		};
		QVERIFY_EXCEPTION_THROWN(
			this->budgetBankLedger_.completedSurvey(
				this->surveyOn_,
				bankAccount,
				expectedReconciled,
				expectedAdded,
				expectedUnusedBudget
			),
			BankError
		);
		QVERIFY_THAT(
			this->budgetBankLedger_.surveys().count(bankAccount),
			equals(static_cast<size_t>(0))
		);

		for (const auto& codeLedgerAccount : this->budgetBankLedger_.generalLedger().ledgers())
		{
			const auto& accountCode{codeLedgerAccount.first};
			const auto& ledgerAccount{codeLedgerAccount.second};

			for (const auto& txnNumberTxn : codeLedgerAccount.second.transactions_)
			{
				const auto& txn{txnNumberTxn.second};
				QVERIFY_THAT(txn.status(), equals(TransactionStatuses::Unreconciled));

				if (accountCode == this->bankAccount_.account_.code_)
					QVERIFY_THAT(ledgerAccount.account_.balance(), equals(originalBalance));
			}
		}
	}

	void bill_completedSurveyNextMonth_ShouldBeReconciled()
	{
		this->budgetBankLedger_.add(this->streaming_);
		const auto firstMonthTxns{
			this->simulateBillPosting(
				1,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn()
			)
		};
		const auto firstMonthSurvey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_,
				40h,
				Days{5}
			)
		};
		vector<TransactionNumber> firstMonthExpectedReconciled{};
		transform(
			firstMonthTxns.cbegin(),
			firstMonthTxns.cend(),
			back_inserter(firstMonthExpectedReconciled),
			[] (const auto& txn)
			{
				return *txn.number();
			}
		);
		firstMonthExpectedReconciled.emplace_back(
			this->budgetBankLedger_.generalLedger().ledger(this->salaryAccount_).transactions_
				.begin()->first
		);
		map<BudgetItemIdentifier, Money> firstMonthExpectedUnusedBudget{
			{
				{this->salaryId_, Money{Iso4217Codes::USD}},
				{this->clothingId_, Money{Iso4217Codes::USD}},
				{this->streamingId_, Money{50, 0u, Iso4217Codes::USD}}
			}
		};
		QVERIFY_NO_THROW(
			this->budgetBankLedger_.completedSurvey(
				this->surveyOn_,
				firstMonthSurvey.reconciled().bankAccount(),
				firstMonthExpectedReconciled,
				{},
				firstMonthExpectedUnusedBudget
			)
		);

		//{ Start the next month's reconciliation.

		const auto txns{
			this->simulateBillPosting(
				1,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn().addMonths(1)
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_.addMonths(1),
				40h,
				Days{5}
			)
		};
		const auto& bankAccount{survey.reconciled().bankAccount()};
		vector<TransactionNumber> expectedReconciled{};
		transform(
			txns.cbegin(),
			txns.cend(),
			back_inserter(expectedReconciled),
			[] (const auto& txn)
			{
				return *txn.number();
			}
		);
		const vector<Transaction> expectedAdded{};
		map<BudgetItemIdentifier, Money> expectedUnusedBudget{
			{
				{this->salaryId_, Money{Iso4217Codes::USD}},
				{this->clothingId_, Money{Iso4217Codes::USD}},
				{this->streamingId_, Money{0, 0u, Iso4217Codes::USD}}
			}
		};
		QVERIFY_NO_THROW(
			this->budgetBankLedger_.completedSurvey(
				this->surveyOn_.addMonths(1),
				bankAccount,
				expectedReconciled,
				expectedAdded,
				expectedUnusedBudget
			)
		);

		//}

		const auto& surveys{this->budgetBankLedger_.surveys().at(bankAccount)};
		const auto& reconciliation{surveys.reconciled()};
		QVERIFY_THAT(
			reconciliation.reconciledOn(),
			equals(Optional<QDate>{this->surveyOn_.addMonths(1)})
		);
		QVERIFY_THAT(reconciliation.balance(), equals(Money{0, 0u, Iso4217Codes::USD}));

		for (const auto& codeLedgerAccount : this->budgetBankLedger_.generalLedger().ledgers())
		{
			const auto& accountCode{codeLedgerAccount.first};
			const auto& ledgerAccount{codeLedgerAccount.second};

			for (const auto& txnNumberTxn : codeLedgerAccount.second.transactions_)
			{
				const auto& txnNum{txnNumberTxn.first};
				const auto& txn{txnNumberTxn.second};

				if (
					find(expectedReconciled.cbegin(), expectedReconciled.cend(), txnNum)
					!= expectedReconciled.cend()
				)
					QVERIFY_THAT(txn.status(), equals(TransactionStatuses::Reconciled));
				else
					QVERIFY_THAT(
						txn.status(),
						anyOf<TransactionStatuses>(
							equals(TransactionStatuses::Unreconciled),
							equals(TransactionStatuses::Reconciled)
						)
					);

				if (accountCode == this->bankAccount_.account_.code_)
				{
					const auto totalUnused{
						sum(
							expectedUnusedBudget.begin()->second.code(),
							expectedUnusedBudget
						)
					};
					QVERIFY_THAT(ledgerAccount.account_.balance(), equals(totalUnused));
				}
			}
		}
	}

	void bill_completedSurveyAddAditional_ShouldBeReconciled()
	{
		this->budgetBankLedger_.add(this->streaming_);
		const auto txns{
			this->simulateBillPosting(
				0,
				this->clothingId_,
				this->clothing_.source(),
				this->clothing_.amount(),
				this->clothing_.nextOccurOn()
			)
		};
		const auto survey{
			this->budgetBankLedger_.projectSurvey(
				this->bankAccount_.bank_,
				this->bankAccount_.account_.code_.number(),
				this->surveyOn_,
				40h,
				Days{5}
			)
		};
		vector<TransactionNumber> expectedReconciled{
			{
				this->budgetBankLedger_.generalLedger().ledger(this->salaryAccount_)
					.transactions_.begin()->first
			}
		};
		const vector<Transaction> expectedAdded{
			{
				{
					this->surveyOn_,
					this->clothingAccount_,
					this->bankAccount_.account_.code_,
					this->clothing_.amount(),
					QStringLiteral("Additional missed transaction.")
				}
			}
		};
		map<BudgetItemIdentifier, Money> expectedUnusedBudget{
			{
				{this->salaryId_, Money{Iso4217Codes::USD}},
				{this->clothingId_, Money{Iso4217Codes::USD}},
				{this->streamingId_, Money{50, 0u, Iso4217Codes::USD}}
			}
		};
		const auto& bankAccount{survey.reconciled().bankAccount()};
		QVERIFY_NO_THROW(
			this->budgetBankLedger_.completedSurvey(
				this->surveyOn_,
				bankAccount,
				expectedReconciled,
				expectedAdded,
				expectedUnusedBudget
			)
		);
		const auto& surveys{this->budgetBankLedger_.surveys().at(bankAccount)};
		const auto& reconciliation{surveys.reconciled()};
		QVERIFY_THAT(reconciliation.reconciledOn(), equals(Optional<QDate>{this->surveyOn_}));
		QVERIFY_THAT(reconciliation.balance(), equals(Money{50, 0u, Iso4217Codes::USD}));

		for (const auto& codeLedgerAccount : this->budgetBankLedger_.generalLedger().ledgers())
		{
			const auto& accountCode{codeLedgerAccount.first};
			const auto& ledgerAccount{codeLedgerAccount.second};

			for (const auto& txnNumberTxn : codeLedgerAccount.second.transactions_)
			{
				const auto& txn{txnNumberTxn.second};
				QVERIFY_THAT(txn.status(), equals(TransactionStatuses::Reconciled));

				if (accountCode == this->bankAccount_.account_.code_)
				{
					const auto totalUnused{
						sum(
							expectedUnusedBudget.begin()->second.code(),
							expectedUnusedBudget
						)
					};
					QVERIFY_THAT(ledgerAccount.account_.balance(), equals(totalUnused));
				}
			}
		}
	}

	//}
};

}}}

QTEST_MAIN(drn::navigation::tests::ProjectedSurveyUT)
#include "ProjectedSurvey-unit-tests.moc"
