/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022                        ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include <pecunia/Codes.h>
using pecunia::currency::Iso4217Codes;
#include <pecunia/Money.h>
using pecunia::currency::Money;
#include <qt-mocks/Matchers.h>
using qt_mocks::equals;
// using QVERIFY_THAT
#include <QDate>
using ::QDate;
#include <QObject>
using ::QObject;
// using Q_OBJECT
#include <QTest>
// using QTEST_MAIN
// using QVERIFY_EXCEPTION_THROWN

#include <map>
using std::map;
#include <set>
using std::set;

#include <accounting/AccountCode.h>
using drn::accounting::AccountCode;
#include <accounting/AccountNumber.h>
using drn::accounting::AccountNumber;
#include <accounting/Transaction.h>
using drn::accounting::Transaction;
using drn::accounting::TransactionNumber;
#include <banking/BankAccount.h>
using drn::banking::BankAccount;
#include <banking/BankAccountType.h>
using drn::banking::BankAccountType;
using drn::banking::SupportedAccountTypes;
#include <banking/BankingErrors.h>
using drn::banking::BankError;
#include <banking/BankName.h>
using drn::banking::BankName;
#include <banking/Reconciliation.h>
using drn::banking::ReconciledBankAccount;
#include <budgeting/BudgetItemTypes.h>
using drn::budgeting::BudgetItemTypes;
#include <budgeting/BudgetItemIdentifier.h>
using drn::budgeting::BudgetItemIdentifier;
#include <budgeting/BudgetSource.h>
using drn::budgeting::BudgetSource;
#include <foundation/Optional.hpp>
using drn::foundation::Optional;
#include <surveying/BankAccountSurvey.h>
using drn::surveying::BankAccountSurvey;


namespace drn
{
namespace surveying
{
namespace tests
{

struct BankAccountSurveyUT : QObject
{
	Q_OBJECT

	const BankAccount bankAccount_{
		BankName{"Home Federal"},
		BankAccountType{
			AccountCode{AccountNumber{1234}, "Personal"},
			SupportedAccountTypes::Chequeing
		}
	};
	const Money balance_{21, 4300u, Iso4217Codes::USD};
	const Optional<QDate> reconciledOn_{QDate::currentDate().addMonths(-1)};
	const ReconciledBankAccount reconciled_{
		this->bankAccount_,
		this->balance_,
		*this->reconciledOn_
	};
	const Money projectedBalance_{1234, 5600u, Iso4217Codes::USD};
	const AccountCode wages_{AccountNumber{4321}, "Job"};
	const map<BudgetItemIdentifier, set<Transaction>> usedBudget_{
		{
			{
				BudgetItemIdentifier{BudgetItemTypes::Wage, BudgetSource{"Job"}},
				{
					{
						{
							this->reconciledOn_->addDays(-1),
							this->bankAccount_.account_.code_,
							this->wages_,
							Money{908, 4400u, Iso4217Codes::USD},
							Optional<QString>{QStringLiteral("Paycheque")},
							TransactionNumber{1}
						}
					}
				}
			}
		}
	};
	map<BudgetItemIdentifier, Money> unusedBudget_{
		{
			{
				BudgetItemIdentifier{BudgetItemTypes::Bill, BudgetSource{"Shudder.com"}},
				Money{4, 7492u, Iso4217Codes::USD}
			}
		}
	};

private slots:
	//{ constructor Tests

	void constructor_default_ShouldBeDefaults()
	{
		const BankAccountSurvey bas{};
		QVERIFY_THAT(bas.reconciled(), equals(ReconciledBankAccount{}));
		QVERIFY_THAT(bas.projectedBalance(), equals(Money{}));
		QVERIFY_THAT(bas.usedBudget(), equals(map<BudgetItemIdentifier, set<Transaction>>{}));
		QVERIFY_THAT(bas.unusedBudget(), equals(map<BudgetItemIdentifier, Money>{}));
	}

	void constructor_fullInit_ShouldBeValues()
	{
		const BankAccountSurvey bas{
			this->reconciled_,
			this->projectedBalance_,
			this->usedBudget_,
			this->unusedBudget_
		};
		QVERIFY_THAT(bas.reconciled(), equals(this->reconciled_));
		QVERIFY_THAT(bas.projectedBalance(), equals(this->projectedBalance_));
		QVERIFY_THAT(bas.usedBudget(), equals(this->usedBudget_));
		QVERIFY_THAT(bas.unusedBudget(), equals(this->unusedBudget_));
	}

	void constructor_bankAccountAndProjectBalanceHaveMismatchedCurrencies_ShouldThrow()
	{
		QVERIFY_EXCEPTION_THROWN(
			(BankAccountSurvey{
				this->reconciled_,
				Money{Iso4217Codes::PLN},
				this->usedBudget_,
				this->unusedBudget_
			}),
			BankError
		);
	}

	//}
};


}}}

QTEST_MAIN(drn::surveying::tests::BankAccountSurveyUT)
#include "BankAccountSurvey-unit-tests.moc"
