/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022                        ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include <pecunia/Codes.h>
using pecunia::currency::Iso4217Codes;
#include <pecunia/Money.h>
using pecunia::currency::Money;
#include <QDate>
using ::QDate;
#include <QObject>
using ::QObject;
// using Q_OBJECT
#include <QTest>
// using QTEST_MAIN
// using QVERIFY_EXCEPTION_THROWN
#include <qt-mocks/Matchers.h>
using qt_mocks::equals;
#include <qt-mocks/Verify.hpp>
// using QVERIFY_THAT

#include <map>
using std::map;

#include <accounting/AccountCode.h>
using drn::accounting::AccountCode;
#include <accounting/AccountingErrors.h>
using drn::accounting::AccountError;
#include <accounting/AccountNumber.h>
using drn::accounting::AccountNumber;
#include <banking/BankAccount.h>
using drn::banking::BankAccount;
#include <banking/BankAccountType.h>
using drn::banking::BankAccountType;
using drn::banking::SupportedAccountTypes;
#include <banking/BankingErrors.h>
using drn::banking::BankError;
#include <banking/BankName.h>
using drn::banking::BankName;
#include <banking/Reconciliation.h>
using drn::banking::ReconciledBankAccount;
#include <budgeting/BudgetingErrors.h>
using drn::budgeting::BudgetItemError;
#include <budgeting/BudgetItemIdentifier.h>
using drn::budgeting::BudgetItemIdentifier;
#include <budgeting/BudgetItemTypes.h>
using drn::budgeting::BudgetItemTypes;
#include <budgeting/BudgetSource.h>
using drn::budgeting::BudgetSource;
#include <foundation/Error.h>
using drn::foundation::Error;
#include <foundation/Optional.hpp>
using drn::foundation::inPlace;
using drn::foundation::Optional;
#include <surveying/ReconciledSurvey.h>
using drn::surveying::ReconciledSurvey;


namespace drn
{
namespace surveying
{
namespace tests
{

struct ReconciledSurveyUT : QObject
{
	Q_OBJECT

	const map<BudgetItemIdentifier, Money> distribution_{
		{
			{
				BudgetItemIdentifier{BudgetItemTypes::Bill, BudgetSource{"Web Hosting"}},
				Money{50, 1234, Iso4217Codes::USD}
			}
		}
	};
	const ReconciledBankAccount reconciled_{
		BankAccount{
			BankName{"C1st Credit Union"},
			BankAccountType{
				AccountCode{AccountNumber{4321}, "Primary Share"},
				SupportedAccountTypes::Savings
			}
		},
		Money{50, 1234, Iso4217Codes::USD},
		Optional<QDate>{inPlace, 2022, 10, 3}
	};
	const Optional<QString> checksum_{
		inPlace,
		QStringLiteral("JCgKc2N7TCsONzuTKBmeEZxz9rwRH7DLBetXA/HcaSs=")
	};

private slots:
	void constructor_DefaultInit_ShouldSet()
	{
		ReconciledSurvey survey{};
		QVERIFY_THAT(survey.distribution(), equals(map<BudgetItemIdentifier, Money>{}));
		QVERIFY_THAT(survey.reconciled(), equals(ReconciledBankAccount{}));
		QVERIFY_FALSE(survey.checksum().hasValue());
	}

	void constructor_PartialInit_ShouldSet()
	{
		const ReconciledSurvey survey{
			this->reconciled_.bankAccount(),
			this->reconciled_.balance().code()
		};
		QVERIFY_THAT(
			survey.reconciled(),
			equals(
				ReconciledBankAccount{
					this->reconciled_.bankAccount(),
					this->reconciled_.balance().code()
				}
			)
		);
		QVERIFY_TRUE(survey.distribution().empty());
		QVERIFY_FALSE(survey.checksum().hasValue());
	}

	void constructor_FullInit_ShouldSet()
	{
		const ReconciledSurvey survey{this->reconciled_, this->distribution_, this->checksum_};
		QVERIFY_THAT(survey.distribution(), equals(this->distribution_));
		QVERIFY_THAT(survey.reconciled(), equals(this->reconciled_));
		QVERIFY_THAT(survey.checksum(), equals(this->checksum_));
	}

	void constructor_ReconciledMissmatchFromValidity_ShouldThrow()
	{
		QVERIFY_EXCEPTION_THROWN(
			(ReconciledSurvey{{}, this->distribution_, this->checksum_}),
			Error
		);
	}

	void constructor_DistributionMissmatchFromValidity_ShouldThrow()
	{
		QVERIFY_EXCEPTION_THROWN(
			(ReconciledSurvey{this->reconciled_, {}, this->checksum_}),
			Error
		);
	}

	void constructor_ValidityMissmatch_ShouldThrow()
	{
		QVERIFY_EXCEPTION_THROWN(
			(ReconciledSurvey{this->reconciled_, this->distribution_, QStringLiteral("123456789")}),
			Error
		);
	}

	//{ Completed Function Tests

	void completed_validValues_ShouldUpdate()
	{
		ReconciledSurvey survey{
			this->reconciled_.bankAccount(),
			this->reconciled_.balance().code()
		};
		survey.completed(
			*this->reconciled_.reconciledOn(),
			this->reconciled_.balance(),
			this->distribution_
		);
		QVERIFY_THAT(survey.distribution(), equals(this->distribution_));
		QVERIFY_THAT(survey.reconciled(), equals(this->reconciled_));
		QVERIFY_THAT(survey.checksum(), equals(this->checksum_));
	}

	void completed_nextValidCompletedSurvey_ShouldUpdate()
	{
		ReconciledSurvey survey{
			this->reconciled_.bankAccount(),
			this->reconciled_.balance().code()
		};
		survey.completed(
			*this->reconciled_.reconciledOn(),
			this->reconciled_.balance(),
			this->distribution_
		);
		QVERIFY_THAT(survey.distribution(), equals(this->distribution_));
		QVERIFY_THAT(survey.reconciled(), equals(this->reconciled_));
		QVERIFY_THAT(survey.checksum(), equals(this->checksum_));
		auto nextReconciled{this->reconciled_};
		nextReconciled.reconcile(
			this->reconciled_.reconciledOn()->addMonths(1),
			this->reconciled_.balance() * 2.0
		);
		auto nextDistribution{this->distribution_};
		nextDistribution.emplace(
			BudgetItemIdentifier{BudgetItemTypes::Bill, BudgetSource{"Domain Registration"}},
			Money{50, 1234, Iso4217Codes::USD}
		);
		survey.completed(
			*nextReconciled.reconciledOn(),
			nextReconciled.balance(),
			nextDistribution
		);
		QVERIFY_THAT(survey.distribution(), equals(nextDistribution));
		QVERIFY_THAT(survey.reconciled(), equals(nextReconciled));
		QVERIFY_THAT(
			survey.checksum(),
			equals(Optional<QString>(inPlace, "7PUXcLSNHjwYPiyh1D1dgCIryDr8W/8PEcowQ3Cy5J4="))
		);
	}

	void completed_balanceDoesNotMatchDistribution_ShouldThrow()
	{
		ReconciledSurvey survey{
			this->reconciled_.bankAccount(),
			this->reconciled_.balance().code()
		};
		QVERIFY_EXCEPTION_THROWN(
			survey.completed(
				*this->reconciled_.reconciledOn(),
				this->reconciled_.balance() * 2.0,
				this->distribution_
			),
			BankError
		);
		QVERIFY_THAT(
			survey.reconciled(),
			equals(
				ReconciledBankAccount{
					this->reconciled_.bankAccount(),
					this->reconciled_.balance().code()
				}
			)
		);
		QVERIFY_TRUE(survey.distribution().empty());
		QVERIFY_FALSE(survey.checksum().hasValue());
	}

	void completed_distributionDoesNotMatchBalance_ShouldThrow()
	{
		ReconciledSurvey survey{
			this->reconciled_.bankAccount(),
			this->reconciled_.balance().code()
		};
		QVERIFY_EXCEPTION_THROWN(
			survey.completed(
				*this->reconciled_.reconciledOn(),
				this->reconciled_.balance(),
				{
					{
						BudgetItemIdentifier{BudgetItemTypes::Bill, BudgetSource{"Web Hosting"}},
						Money{250, 1234, Iso4217Codes::USD}
					}
				}
			),
			BankError
		);
		QVERIFY_THAT(
			survey.reconciled(),
			equals(
				ReconciledBankAccount{
					this->reconciled_.bankAccount(),
					this->reconciled_.balance().code()
				}
			)
		);
		QVERIFY_TRUE(survey.distribution().empty());
		QVERIFY_FALSE(survey.checksum().hasValue());
	}

	void completed_nextSurveyInPast_ShouldThrow()
	{
		ReconciledSurvey survey{
			this->reconciled_.bankAccount(),
			this->reconciled_.balance().code()
		};
		survey.completed(
			*this->reconciled_.reconciledOn(),
			this->reconciled_.balance(),
			this->distribution_
		);
		QVERIFY_THAT(survey.distribution(), equals(this->distribution_));
		QVERIFY_THAT(survey.reconciled(), equals(this->reconciled_));
		QVERIFY_THAT(survey.checksum(), equals(this->checksum_));
		const ReconciledBankAccount nextReconciled{
			this->reconciled_.bankAccount(),
			Money{100, 2468, Iso4217Codes::USD},
			Optional<QDate>{inPlace, 2022, 9, 3}
		};
		auto nextDistribution{this->distribution_};
		nextDistribution.emplace(
			BudgetItemIdentifier{BudgetItemTypes::Bill, BudgetSource{"Domain Registration"}},
			Money{50, 1234, Iso4217Codes::USD}
		);
		QVERIFY_EXCEPTION_THROWN(
			survey.completed(
				*nextReconciled.reconciledOn(),
				nextReconciled.balance(),
				nextDistribution
			),
			BankError
		);
		QVERIFY_THAT(survey.distribution(), equals(this->distribution_));
		QVERIFY_THAT(survey.reconciled(), equals(this->reconciled_));
		QVERIFY_THAT(survey.checksum(), equals(this->checksum_));
	}

	//}

	//{ Renamed Function Tests

	void renamed_newBankName_ShouldUpdateAndRecalculateChecksum()
	{
		ReconciledSurvey survey{this->reconciled_, this->distribution_, this->checksum_};
		const BankName newName{"New Bank"};
		survey.renamed(this->reconciled_.bankAccount().bank_, newName);
		QVERIFY_THAT(survey.distribution(), equals(this->distribution_));
		const ReconciledBankAccount expectedReconciled{
			BankAccount{newName, this->reconciled_.bankAccount().account_},
			this->reconciled_.balance(),
			this->reconciled_.reconciledOn()
		};
		QVERIFY_THAT(survey.reconciled(), equals(expectedReconciled));
		QVERIFY_THAT(
			survey.checksum(),
			equals(Optional<QString>{"ZItCRNuY+e8OIvW4fUc8FznTKAJJSyDWrt/hGfYsFQY="})
		);
	}

	void renamed_bankNameDoesNotExist_ShouldThrow()
	{
		ReconciledSurvey survey{this->reconciled_, this->distribution_, this->checksum_};
		const BankName newName{"New Bank"};
		QVERIFY_EXCEPTION_THROWN(
			survey.renamed(BankName{"Does Not Exist"}, newName),
			BankError
		);
		QVERIFY_THAT(survey.distribution(), equals(this->distribution_));
		QVERIFY_THAT(survey.reconciled(), equals(this->reconciled_));
		QVERIFY_THAT(survey.checksum(), equals(this->checksum_));
	}

	void renamed_newAccountCode_ShouldUpdateAndRecalculateChecksum()
	{
		ReconciledSurvey survey{this->reconciled_, this->distribution_, this->checksum_};
		const AccountCode change{AccountNumber{10101u}, "New Account Code"};
		survey.renamed(this->reconciled_.bankAccount().account_.code_.number(), change);
		QVERIFY_THAT(survey.distribution(), equals(this->distribution_));
		const ReconciledBankAccount expectedReconciled{
			BankAccount{
				this->reconciled_.bankAccount().bank_,
				BankAccountType{change, this->reconciled_.bankAccount().account_.type_}
			},
			this->reconciled_.balance(),
			this->reconciled_.reconciledOn()
		};
		QVERIFY_THAT(survey.reconciled(), equals(expectedReconciled));
		QVERIFY_THAT(
			survey.checksum(),
			equals(Optional<QString>{"h3/KhvDCNLZCUMS3z72JKmANgKUeCOza0piv8dPpmws="})
		);
	}

	void renamed_accountNumberDoesNotExist_ShouldThrow()
	{
		ReconciledSurvey survey{this->reconciled_, this->distribution_, this->checksum_};
		QVERIFY_EXCEPTION_THROWN(
			survey.renamed(AccountNumber{12345678}, this->reconciled_.bankAccount().account_.code_),
			AccountError
		);
		QVERIFY_THAT(survey.distribution(), equals(this->distribution_));
		QVERIFY_THAT(survey.reconciled(), equals(this->reconciled_));
		QVERIFY_THAT(survey.checksum(), equals(this->checksum_));
	}

	void renamed_newBudgetName_ShouldUpdateAndRecalculateChecksum()
	{
		ReconciledSurvey survey{this->reconciled_, this->distribution_, this->checksum_};
		const BudgetSource newName{"New Budget"};
		survey.renamed(this->distribution_.cbegin()->first, newName);
		const map<BudgetItemIdentifier, Money> expectedDistribution{
			{
				{
					BudgetItemIdentifier{this->distribution_.cbegin()->first.type_, newName},
					this->distribution_.cbegin()->second
				}
			}
		};
		QVERIFY_THAT(survey.distribution(), equals(expectedDistribution));
		QVERIFY_THAT(survey.reconciled(), equals(this->reconciled_));
		QVERIFY_THAT(
			survey.checksum(),
			equals(Optional<QString>{"khJIYCe4PUXzWg+65BNBcsDzjO8350f4rmHxt6iXhAs="})
		);
	}

	void renamed_budgetNameDoesNotExist_ShouldThrow()
	{
		ReconciledSurvey survey{this->reconciled_, this->distribution_, this->checksum_};
		const BudgetSource newName{"New Budget"};
		QVERIFY_EXCEPTION_THROWN(
			survey.renamed(BudgetItemIdentifier{BudgetItemTypes::Goal, BudgetSource{"T"}}, newName),
			BudgetItemError
		);
		QVERIFY_THAT(survey.distribution(), equals(this->distribution_));
		QVERIFY_THAT(survey.reconciled(), equals(this->reconciled_));
		QVERIFY_THAT(survey.checksum(), equals(this->checksum_));
	}

	//}
};

}}}

QTEST_MAIN(drn::surveying::tests::ReconciledSurveyUT)
#include "ReconciledSurvey-unit-tests.moc"
